#! /usr/bin/perl
#
# Based on requested URI, run the query and return the 
# results in XML.
#
use strict;
use warnings;
use CGI qw/:standard/;
use CGI::Carp qw/fatalsToBrowser/;
use DBI;
use XML::Generator::DBI;
use XML::SAX::Writer;
use Data::Dumper;
require 'common.pl';

my %valid_conditions=(
    'NEWER_THAN_DAYS' => ' > dateadd(dd, ? * -1, getdate())',
    'OLDER_THAN_DAYS' => ' < dateadd(dd, ? * -1, getdate())',
    '=' => '= ?',
    'ISNOT' => '!= ?',
);

#
# get system configuration vars from <scriptName>.conf
#
my $confFile = $0;
$confFile =~ s/\.(cgi|pl)$/.conf/;
{ package sysConf; do "$confFile"
        or die "Can't read configuration file: $confFile"}

my $q = new CGI;

my $script = $ARGV[0] || &script_name;
$script=~s/^\///; # knock off the first slash, so we don't get empty array elements.
$script=~s/\/\?.*//; # knock off any query strings (starts with /?)
$script = $q->param('query') if $q->param('query');
my @requestURL = split('/',$script);

my $scriptname=shift @requestURL;
my $db=shift(@requestURL) || '';
my $table=shift(@requestURL) || '';

my $valid_columns; # global

if(exists($sysConf::db->{$db}) && valid_table($db, $table)) {
    my $dbh = connectDB($db);
    if($q->param('show_columns')) {
        my $qstr=getColumnsSql($table);
        getXml($dbh, $qstr, []);
    } else {
        $valid_columns = validColumns($dbh, $table);
        printTableData($dbh, $table, \@requestURL);
    }
} else {
    #print "$script -> $db, $table\n";
    showUsage("$script -> $db, $table\n");
}

exit;

sub valid_table() {
    # is this a valid table for given db?
    my $db=shift;
    my $table=shift;
    my $conf=$sysConf::db->{$db};
    return unless defined $conf;
    return unless defined $conf->{'tables'};
    my %tables=();
    foreach (@{$conf->{'tables'}}) {
        $tables{$_} = 1;
    }
    return exists($tables{$table});
}

sub printTableData {
    my $dbh = shift;
    my $table=shift;
    my $params=shift;

    my $qstr;
    my $select='';
    my @distinct=$q->param('distinct');
    my ($filter, $bindparams)  = buildFilter($params);

    if($#distinct >= 0) {
        my @d=();
        foreach my $d (@distinct) {
            if(exists $valid_columns->{$d} ) {
                push @d, $d;
            }
        }
        $select = " DISTINCT (" . join(",", @d) . ") ";
    } else {
        $select='*';
    }

    if ($q->param("count")) {
        $qstr = "SELECT COUNT($select) as 'count' FROM $table $filter ";
    } else {
        $qstr = "SELECT $select FROM $table $filter ";
    }

    if($q->param('show_columns')) {
        $qstr=getColumnsSql($table);
    }
    my $limit=$q->param('limit');
    $limit=~s/[^0-9]//g;
    if($limit) {
        $dbh->do('set rowcount ' . $limit);
    }

    return getXml($dbh, $qstr, $bindparams);
}

sub getXml() {
    my $dbh=shift;
    my $qstr=shift;
    my $bindparams=shift;
    print "Content-type:  text/xml\n\n";
    print '<?xml version="1.0" ?>' . "\n";
    print '<!-- ' . $qstr . " -->\n";
    print '<!-- ' . Dumper($bindparams) . " -->\n";
    my $sth = $dbh->prepare($qstr) || die $DBI::errstr;
#    $sth->execute($bindparams) || die $DBI::errstr;
#    while(my $r = $sth->fetchrow_hashref()) {
#        print keys(%$r);
#        print "\n";
#    }
#    return;
    my $generator = XML::Generator::DBI->new(
        Handler => XML::SAX::Writer->new(EncodeFrom => 'ISO-8859-1'),
        Indent => 1,
        RootElement=>'Space',
        dbh => $dbh,
    );
    $generator->execute($sth, $bindparams);
    $dbh->do('set rowcount 0');
}


#
# buildFilter build a 'WHERE' clause from the CGI parms
#
sub buildFilter {
    my $requestParams=shift;
    my @qstring = ();
    my @bindparams = ();
    #my %inparams=();
    my %inparams=('in' => {}, 'notin' => {});

    for(my $i=0; $i <= $#$requestParams; $i+=3 ) {
        my $col = $requestParams->[$i];
        my $cond = $requestParams->[$i+1];
        my $val = $requestParams->[$i+2];
        unless(exists $valid_columns->{$col}) {
            next;
        }
        #print "*$cond*\n";
        #print "*$valid_conditions{$cond}*\n";
        if(exists $valid_conditions{$cond}) {
            # Known condition type, use the canned SQL.
            push @qstring, "($col " . $valid_conditions{$cond} . ')';
            push @bindparams, $val;
        } else {
            # Assume list of values.
            # Push the value onto the column's array in the inparams hash.
            # Later it will be pulled back in an orderly fashion to produce a WHERE ... IN clause.
            if($cond eq 'NOTIN') {
                $inparams{'notin'}->{$col} = [] unless exists $inparams{'notin'}->{$col};
                push @{ $inparams{'notin'}->{$col} }, $val;
            } else {
                $inparams{'in'}->{$col} = [] unless exists $inparams{'in'}->{$col};
                push @{ $inparams{'in'}->{$col} }, $val;
            }
        }
    }
    foreach my $type ('in','notin') {
        foreach my $k (keys %{ $inparams{$type} }) {
            my @qmarks=();
            for my $val (@{ $inparams{$type}->{$k} }) {
                push @qmarks, '?';
                push @bindparams, $val;
            }
            my $clause = ' IN ';
            $clause = 'NOT IN' if $type eq 'notin';
            push @qstring, "($k $clause (" . join(',', @qmarks) . '))';
        }
    }

    my $qstring = join(' AND ', @qstring);
    if ($qstring) {
        $qstring = "WHERE $qstring";
    }

    return $qstring, \@bindparams;
}

sub validColumns() {
    my $dbh  =shift;
    my $table=shift;
    my $qstr = getColumnsSql($table);
    my $sth=$dbh->prepare($qstr);
    $sth->execute;
    my $hash_ref = $sth->fetchall_hashref('name');
    return $hash_ref;
}

sub getColumnsSql() {
    my $table=shift;
    "SELECT * FROM syscolumns WHERE id=object_id('$table')";
}

sub showUsage() {
    print $q->header(-content_type => 'text/html', -status => '200');
    #print "Content-type: text/plain\n\n";
    print<<EOF;
    <html>
    <head>
    <title>
    XML interface Help
    </title>
    <body>
    <p>
    @_
    </p>
<dl>
<dt># /spaceRest/&lt;dbname&gt;/foo/?distinct=bar&distinct=baz</dt>
  <dd>select distinct (bar,baz) from foo;</dd>
<dt># /spaceRest/&lt;dbname&gt;/foo/bar/IN/baz/bar/IN/quux/fred/=/wilma</dt>
  <dd>select * from foo where bar in (baz,quux) and fred=wilma;</dd>
<dt># /spaceRest/&lt;dbname&gt;/foo/bar/=/baz/?count=1</dt>
  <dd>select count(*) from foo where bar=baz;</dd>
<dt># /spaceRest/&lt;dbname&gt;/foo?show_columns=1</dt>
  <dd>(show column details for foo table);</dd>
</dl>

  <p>
count and distinct can be combined. /?limit=n can also be used to limit the results to n rows.
</p>
<p>
Condition codes (IN is default):
</p>
<table border="1">
<tr>
<th>URL fragment</th>
<th>SQL</th>
</tr>
<tr>
<td>/Column/IN/value/column/IN/value2</td>
<td>
Column IN (value, value2)
</td>
</tr>
EOF

    foreach my $k (keys %valid_conditions) {
        print "<tr>";
        print "<td>/Column/$k/Value </td><td>Column $valid_conditions{$k}</td>\n";
        print "</tr>";
    }

    print "</table>\n<p>\nAvailable tables:\n</p>\n<dl>";
    foreach my $db (keys %{ $sysConf::db }) {
        print "<dt>Database: $db</dt>\n";
        foreach my $tables ($sysConf::db -> {$db}->{'tables'}) {
            print "<dd>  Tables -> " . join(', ',sort @$tables) . "</dd>\n";
        }
    }
    print<<EOF;
</dl>
</body>
</html>
EOF

}
